package edu.wpi.banmqp.basestation;

import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.UUID;

import android.annotation.SuppressLint;
import android.app.Activity;
import android.app.AlertDialog;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothSocket;
import android.content.DialogInterface;
import android.content.Intent;
import android.os.AsyncTask;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.widget.ArrayAdapter;
import android.widget.TextView;
import android.widget.Toast;
import de.congrace.exp4j.Calculable;
import edu.wpi.banmqp.shimmer.BasicShimmerMote;
import edu.wpi.banmqp.shimmer.IShimmer;
import edu.wpi.banmqp.shimmer.commands.learned.ILearnedCommand;
import edu.wpi.banmqp.shimmer.commands.learned.ParamMapping;
import edu.wpi.banmqp.shimmer.data.ValueMapping;
import edu.wpi.banmqp.shimmer.sensor.ISensor;
import edu.wpi.banmqp.util.BANLog;
import edu.wpi.banmqp.util.ColorQueue;

@SuppressLint("UseSparseArrays")
@SuppressWarnings("rawtypes")
public class MainActivity extends Activity implements CommandDialogFragment.OnCompleteListener {

	/** The settings menu. */
	private final int SETTINGS_MENU = 0;

	private static DecimalFormat df = new DecimalFormat("#.####");


	private static HashMap<Integer, TextView> sensorNames;
	private static HashMap<Integer, TextView> sensorValues;
	private static HashMap<ISensor, ArrayList<TextView>> sensorsToViews;
	private static HashMap<TextView, ISensor> viewToSensor;
	private static ArrayList<Integer> viewsUsed;
	private static ArrayList<AsyncTask> runningThreads;
	private static int nextViewToUse = 0;
	private static int maxIndexOfViews = 11;
	private static ArrayList<IShimmer> connectedDevices;
	private static ColorQueue colors;
	private static Handler graphHandler = null;
	private static String addrOfMoteToGraph = "";

	/** The Constant defaultUUID. */
	public static final UUID defaultUUID = UUID
			.fromString("00001101-0000-1000-8000-00805F9B34FB");
	/** The Constant mHandler. */
	private static final Handler dataHandler = new Handler() {
		@Override
		public void handleMessage(Message msg) {
			BANLog.v("HANDLER", "Message recieved");
			int messageType = msg.getData().getInt("MESSAGE_TYPE");
			int sensorID;
			String addr;
			IShimmer mote;
			ISensor sensor;
			switch(messageType){
			// Pull stuff out of data packet, update any relevent views
			case MESSAGE_TYPE.DATA:
				sensorID = msg.getData().getInt("SensorID");
				long[] values = msg.getData().getLongArray("values");
				addr = msg.getData().getString("Address");
				mote = getMoteByAddr(addr);
				if(mote == null){
					BANLog.e("PROCESS_DATA", "Unable to find mote with addr " + addr);
					return;
				}
				sensor = mote.getSensorById(sensorID);
				ArrayList<TextView> views = sensorsToViews.get(sensor);
				if(views == null){
					BANLog.e("PROCESS_DATA", "Unable to find any views for mote " + addr + " sensor " + sensorID);
					return;
				}
				ArrayList<ValueMapping> vm = sensor.getValueMappings();
				// Iterate over the views, update each with new data value
				for(int i = 0; i < views.size(); i++){
					if(i >= values.length){
						BANLog.d("PROCESS_DATA", "Ignoring extra data values provided for mote " + addr + " sensor " + sensorID);
						break;
					}
					try{
						Calculable ce = vm.get(i).getConversionExpression();
						ce.setVariable("x", values[i]);
						double value = ce.calculate();
						views.get(i).setText(df.format(value) + " " + vm.get(i).getUnits());
						if(sensorID == 1 && graphHandler != null && addrOfMoteToGraph.equals(addr)){
							Message m = graphHandler.obtainMessage();
							Bundle b = m.getData();
							b.putInt("index", i);
							b.putInt("numValues", views.size());
							b.putDouble("data", value);
							b.putString("deviceName", sensor.toString() + " on " + addr.replaceAll(":", "").substring(8));
							b.putString("units", vm.get(i).getUnits());
							b.putString("legendTitle", vm.get(i).getName());
							m.sendToTarget();
						}

					}catch(NullPointerException e){
						BANLog.d("PROCESS_DATA", "Unable to find valuemapping for " + addr + " sensor " + sensorID);
						return;
					}



				}
				break;
			case MESSAGE_TYPE.SENSOR_ADD:
				sensorID = msg.getData().getInt("SensorID");
				addr = msg.getData().getString("Address");
				mote = getMoteByAddr(addr);
				String lastFour = addr.replaceAll(":", "").substring(8);
				if(mote == null){
					BANLog.e("PROCESS_SENSOR_ADD", "Unable to find mote with addr " + addr);
					return;
				}
				BANLog.d("UPDATE_GUI", "Attempting to bind sensor values to next availible GUI fields");
				sensor = mote.getSensorById(sensorID);
				ArrayList<ValueMapping> valueMappings = sensor.getValueMappings();
				int color = colors.nextColor();
				//Try to map each value to a field
				for(ValueMapping v : valueMappings){
					BANLog.d("UPDATE_GUI", "Looking for view for " + v.getName());
					// Sanity check to make sure we aren't going to re-use a view or go out of bounds
					if(!viewsUsed.contains(nextViewToUse) && nextViewToUse <= maxIndexOfViews){
						// Update viewlist associated with this sensor
						ArrayList<TextView> viewList = sensorsToViews.get(sensor);
						if(viewList == null){
							viewList = new ArrayList<TextView>();
						}
						TextView valueView = sensorValues.get(nextViewToUse);
						viewList.add(valueView);
						valueView.setVisibility(View.VISIBLE);
						valueView.setBackgroundColor(color);
						sensorsToViews.put(sensor, viewList);
						// Update displayed name of sensor
						TextView nameView = sensorNames.get(nextViewToUse);
						nameView.setText(sensor.toString() + ":\n" + v.getName() + " on " + lastFour);
						nameView.setClickable(true);
						nameView.setBackgroundColor(color);
						nameView.setVisibility(View.VISIBLE);
						viewToSensor.put(nameView, sensor);
						BANLog.d("UPDATE_GUI", "Added view for " + v.getName());
						viewsUsed.add(nextViewToUse);
						nextViewToUse++;


					}

					else{
						BANLog.e("UPDATE_GUI", "Insufficient views for " + v.getName());
					}

				}
				break;
			case MESSAGE_TYPE.REDRAW:
				// Record the motes to restart streaming
				for(IShimmer s : connectedDevices){
					if(s.isStreaming()){
						s.sendMoteData();
						s.resumeStreaming(true);
					}
				}
				// Reset association between sensors and views
				sensorsToViews = new HashMap<ISensor, ArrayList<TextView>>();
				viewToSensor = new HashMap<TextView, ISensor>();
				viewsUsed = new ArrayList<Integer>();
				nextViewToUse = 0;
				colors = new ColorQueue();
				// Hide all views
				for(TextView tv : sensorValues.values()){
					tv.setVisibility(View.INVISIBLE);
				}
				for(TextView tv : sensorNames.values()){
					tv.setVisibility(View.INVISIBLE);
				}
				// Reload everything we know about the motes
				for(IShimmer s: connectedDevices){
					s.sendSensorInquiry(true);
				}
				break;



			}





			BANLog.v("HANDLER", "Message processed");
		}

	};

	/*
	 * (non-Javadoc)
	 * 
	 * @see android.app.Activity#onCreate(android.os.Bundle)
	 */
	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		setContentView(R.layout.activity_main);
		deviceCompatabilityChecks();
		initializeValues();
		BANLog.d("LAUNCH", "Initilization Complete");

	}



	protected static IShimmer getMoteByAddr(String addr) {
		for(IShimmer i : connectedDevices){
			if(i.getBluetoothAddr().equalsIgnoreCase(addr)){
				return i;
			}
		}
		return null;
	}



	/**
	 * Initialize values.
	 */
	private void initializeValues() {
		sensorNames = new HashMap<Integer, TextView>();
		sensorValues = new HashMap<Integer, TextView>();
		sensorNames.put(0, (TextView) findViewById(R.id.sensorname0textview));
		sensorNames.put(1, (TextView) findViewById(R.id.sensorname1textview));
		sensorNames.put(2, (TextView) findViewById(R.id.sensorname2textview));
		sensorNames.put(3, (TextView) findViewById(R.id.sensorname3textview));
		sensorNames.put(4, (TextView) findViewById(R.id.sensorname4textview));
		sensorNames.put(5, (TextView) findViewById(R.id.sensorname5textview));
		sensorNames.put(6, (TextView) findViewById(R.id.sensorname6textview));
		sensorNames.put(7, (TextView) findViewById(R.id.sensorname7textview));
		sensorNames.put(8, (TextView) findViewById(R.id.sensorname8textview));
		sensorNames.put(9, (TextView) findViewById(R.id.sensorname9textview));
		sensorNames.put(10, (TextView) findViewById(R.id.sensorname10textview));
		sensorNames.put(11, (TextView) findViewById(R.id.sensorname11textview));
		sensorValues.put(0, (TextView) findViewById(R.id.sensorvalue0textview));
		sensorValues.put(1, (TextView) findViewById(R.id.sensorvalue1textview));
		sensorValues.put(2, (TextView) findViewById(R.id.sensorvalue2textview));
		sensorValues.put(3, (TextView) findViewById(R.id.sensorvalue3textview));
		sensorValues.put(4, (TextView) findViewById(R.id.sensorvalue4textview));
		sensorValues.put(5, (TextView) findViewById(R.id.sensorvalue5textview));
		sensorValues.put(6, (TextView) findViewById(R.id.sensorvalue6textview));
		sensorValues.put(7, (TextView) findViewById(R.id.sensorvalue7textview));
		sensorValues.put(8, (TextView) findViewById(R.id.sensorvalue8textview));
		sensorValues.put(9, (TextView) findViewById(R.id.sensorvalue9textview));
		sensorValues.put(10, (TextView) findViewById(R.id.sensorvalue10textview));
		sensorValues.put(11, (TextView) findViewById(R.id.sensorvalue11textview));
		sensorsToViews = new HashMap<ISensor, ArrayList<TextView>>();
		viewToSensor = new HashMap<TextView, ISensor>();
		viewsUsed = new ArrayList<Integer>();
		connectedDevices = new ArrayList<IShimmer>();
		runningThreads = new ArrayList<AsyncTask>();
		colors = new ColorQueue();
	}

	/**
	 * Device compatibility checks.
	 */
	private void deviceCompatabilityChecks() {
		BluetoothAdapter ba = BluetoothAdapter.getDefaultAdapter();
		if (ba == null) {
			Toast.makeText(getApplicationContext(),
					"Bluetooth is not supported by this device, exiting",
					Toast.LENGTH_SHORT).show();
			finish();
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see android.app.Activity#onCreateOptionsMenu(android.view.Menu)
	 */
	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		// Inflate the menu; this adds items to the action bar if it is present.
		getMenuInflater().inflate(R.menu.main, menu);
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see android.app.Activity#onOptionsItemSelected(android.view.MenuItem)
	 */
	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		switch (item.getItemId()) {
		case R.id.action_settings:
			Intent i = new Intent(this, SettingsActivity.class);
			i.setAction(Intent.ACTION_MAIN);
			startActivityForResult(i, SETTINGS_MENU);
			return true;
		case R.id.action_log:
			Intent j = new Intent(this, LogActivity.class);
			j.setAction(Intent.ACTION_MAIN);
			startActivity(j);
			return true;
		case R.id.action_startstream:
			for(IShimmer s: connectedDevices){
				s.sendMoteData();
			}
			return true;
		case R.id.PlotInit:
			showGraphDialog();
		}


		return false;
	}

	private void showGraphDialog() {
		
		ArrayList<String> deviceNames = new ArrayList<String>(); 
		for (IShimmer s: connectedDevices) {
			deviceNames.add(s.getBluetoothAddr());
		}
		AlertDialog.Builder builderSingle = new AlertDialog.Builder(
                MainActivity.this);
        builderSingle.setIcon(R.drawable.ic_launcher);
        builderSingle.setTitle("Select Value:");
        final ArrayAdapter<String> arrayAdapter = new ArrayAdapter<String>(
                MainActivity.this,
                android.R.layout.select_dialog_singlechoice);
        arrayAdapter.addAll(deviceNames);
        builderSingle.setNegativeButton("Cancel",
                new DialogInterface.OnClickListener() {
                    @Override
                    public void onClick(DialogInterface dialog, int which) {
                        dialog.dismiss();
                        return;
                    }
                });
        builderSingle.setAdapter(arrayAdapter,
                new DialogInterface.OnClickListener() {
                    @Override
                    public void onClick(DialogInterface dialog, int which) {
                    	addrOfMoteToGraph = arrayAdapter.getItem(which);
                        graphHandler = PlotActivity.getHandler();
                        Intent k = new Intent(getApplicationContext(), PlotActivity.class);
                        k.setAction(Intent.ACTION_MAIN);
                        startActivity(k);
                    }
				
                });
        builderSingle.show();
		
	}



	/*
	 * (non-Javadoc)
	 * 
	 * @see android.app.Activity#onActivityResult(int, int,
	 * android.content.Intent)
	 */
	@Override
	protected void onActivityResult(int requestCode, int resultCode, Intent data) {
		if (resultCode == Activity.RESULT_CANCELED) {
		} else if (resultCode == Activity.RESULT_OK
				&& requestCode == SETTINGS_MENU) {

			BANLog.d("MAIN_ACTIVITY", "Returning to main");
			String btid = data.getExtras().getString("BTID");
			BANLog.d("MAIN_ACTIVITY", "Got BTID " + btid);
			boolean newDevice = true;
			IShimmer bsm = null;
			for(IShimmer s : connectedDevices){
				if(s.getBluetoothAddr().equalsIgnoreCase(btid)){
					newDevice = false;
					bsm = s;
				}
			}
			if(newDevice){
				BluetoothSocket sock = null;

				try {
					sock = BluetoothAdapter.getDefaultAdapter()
							.getRemoteDevice(btid)
							.createRfcommSocketToServiceRecord(defaultUUID);
					bsm = new BasicShimmerMote(sock);
					connectedDevices.add(bsm);
				} catch (IOException e) {
					BANLog.e("DEVICE_CONNECT", "Error connecting to device at addr " + btid);
					e.printStackTrace();
				}


			}
			if(bsm != null){
				MoteThread RD = new MoteThread(dataHandler, bsm);
				runningThreads.add(RD.executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR,
						new Void[0]));

			}
		}


	}

	public void showCustomCommands(View view){
		BANLog.d("CUSTOM_COMMAND", "Trying to show list of custom commands");
		ISensor s = viewToSensor.get((TextView) view);
		if(s != null){
			ILearnedCommand[] cmds = s.getLearnedCommands().toArray(new ILearnedCommand[0]);
			BANLog.d("CUSTOM_COMMAND", "CMD LIST: " + Arrays.toString(cmds));
			CommandDialogFragment dialog = new CommandDialogFragment(cmds);
			dialog.show(getFragmentManager(), "dialog");
		}
	}



	@Override
	public void onComplete(final ILearnedCommand iLearnedCommand) {
		BANLog.d("CUSTOM_COMMAND", "User wants to run command " + iLearnedCommand.getName() + " on sensor " + iLearnedCommand.getSensor().getID());
		ArrayList<ParamMapping> params = iLearnedCommand.getParams();
		for(ParamMapping p : params){
			if(p.getRestrictionSet().contains(",")){
				final int valueSize = p.getSizeInBytes();
				String[] possibleValues = p.getRestrictionSet().split(",");
				AlertDialog.Builder builderSingle = new AlertDialog.Builder(
	                    MainActivity.this);
	            builderSingle.setIcon(R.drawable.ic_launcher);
	            builderSingle.setTitle("Select Value:");
	            final ArrayAdapter<String> arrayAdapter = new ArrayAdapter<String>(
	                    MainActivity.this,
	                    android.R.layout.select_dialog_singlechoice);
	            arrayAdapter.addAll(possibleValues);
	            builderSingle.setNegativeButton("Cancel",
	                    new DialogInterface.OnClickListener() {
	                        @Override
	                        public void onClick(DialogInterface dialog, int which) {
	                            dialog.dismiss();
	                            return;
	                        }
	                    });
	            builderSingle.setAdapter(arrayAdapter,
	                    new DialogInterface.OnClickListener() {
	                        @Override
	                        public void onClick(DialogInterface dialog, int which) {
	                            byte[] valueArray = new byte[valueSize];
	                            valueArray[valueArray.length - 1] = (byte) which;
	                            iLearnedCommand.getParamBytes().add(valueArray);
	                            BANLog.d("CUSTOM_COMMAND", "Adding value " + Arrays.toString(valueArray) + " for custom command");
	                            checkIfAllParamsAdded(iLearnedCommand);
	                            
	                        }
						
	                    });
	            builderSingle.show();
			}
		}		
	}
	
	private void checkIfAllParamsAdded(
			ILearnedCommand iLearnedCommand) {
		if(iLearnedCommand.getParamBytes().size() == iLearnedCommand.getParams().size()){
			int totalNumBytes = 0;
			BANLog.d("CUSTOM_COMMAND", "TEST " + iLearnedCommand.getParamBytes().size());
			for(byte[] b : iLearnedCommand.getParamBytes()){
				totalNumBytes = totalNumBytes + b.length;
				BANLog.d("CUSTOM_COMMAND", "Increased allocation for params by " + b.length + " bytes");
			}
			// Shove them all together
			byte[] bValues = new byte[totalNumBytes];
			BANLog.d("CUSTOM_COMMAND", "Allocated " + totalNumBytes + " to hold params");
			int bValueIndex = 0;
			for(byte[] bArray : iLearnedCommand.getParamBytes()){
				for(byte b : bArray){
					if(bValueIndex <= bValues.length){
						bValues[bValueIndex++] = b;
					}
					else{
						BANLog.e("CUSTOM_COMMAND", "Size mismatch!");
						break;
					}
				}
			}
			// Instruct the mote to send the command
			iLearnedCommand.getSensor().getMote().sendCustomCommand(iLearnedCommand.getSensor().getID(), iLearnedCommand.getID(), bValues);
			// Clear the stored bytes
			iLearnedCommand.getParamBytes().clear();
		}
		
	}
	
}
