package edu.wpi.banmqp.basestation;

import java.util.Date;

import android.os.AsyncTask;
import android.os.Handler;
import edu.wpi.banmqp.shimmer.IShimmer;
import edu.wpi.banmqp.util.BANLog;

/**
 * Used to read data from an IShimmer device
 */
public class MoteThread extends AsyncTask<Void, Void, Void> {

	private Handler h;
	private IShimmer mote;
	private long lastSend;
	private long timeout = 3000; // 3 seconds


	/**
	 * 
	 * @param h handler that will parse the recieved data
	 * @param mote the device to read data from
	 */
	public MoteThread(Handler h, IShimmer mote) {
		this.h = h;
		this.mote = mote;
		lastSend = new Date().getTime();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see android.os.AsyncTask#doInBackground(Params[])
	 */
	@Override
	protected Void doInBackground(Void... params) {
		// Make sure we connect
		while(!mote.isConnected()){}
		// Learn everything we can about the mote
		mote.sendSensorInquiry(true);
		
		// Main processing loop
		while(true){
			//Check if we can send. If we have reached the timeout, just send anyway
			if((mote.clearToSend() && mote.waitingToSend()) || ((new Date().getTime()) - lastSend >= timeout)){
				mote.sendNextMsg();
				lastSend = new Date().getTime();
			}
			// Process any incoming packet
			if(mote.bytesAvailible() >= 5){
				BANLog.d("PACKET_RECIEVED", "New packet recieved from " + mote.getBluetoothAddr());
				mote.processPacket(h);
			}
		}
	}
	

	/**
	 * @return the mote
	 */
	public IShimmer getMote() {
		return mote;
	}



}


